#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <conio.h>
#include "wvga.h"
#include "texture.h"
#include "palman.h"

#ifndef M_PI
#define M_PI	3.14159265
#endif

#define WIDTH	320
#define HEIGHT	200

int init(void);
void cleanup(void);
void display(void);
int calc_tunnel_mapping(void);

static unsigned char *fbuf;
static struct texture *tex;
static int colrange;

static char *tex_fname = "data/wall2.ppm";

static unsigned long *umap, *vmap;

static unsigned long frames;
static unsigned long start_time;

int main(int argc, char **argv)
{
	if(argc > 1) {
		tex_fname = argv[1];
	}

	if(init() == -1) {
		return 1;
	}

	while(!kbhit()) {
		display();
	}
	cleanup();
	return 0;
}

int init(void)
{
	int i;
	struct palm_color *pal;

	if(!(fbuf = malloc(WIDTH * HEIGHT))) {
		fprintf(stderr, "failed to allocate framebuffer\n");
		return -1;
	}

	if(calc_tunnel_mapping() == -1) {
		fprintf(stderr, "failed to precalc tunnel mapping\n");
		return 1;
	}

	if(tex_fname) {
		if(!(tex = load_texture(tex_fname))) {
			fprintf(stderr, "failed to load texture: %s\n", tex_fname);
			free(fbuf);
			return -1;
		}

	} else {
		fprintf(stderr, "FOO\n");
		return -1;
	}

	set_video_mode(0x13);

	palm_build();
	get_texture_pixels(tex);

	pal = palm_palette();
	for(i=0; i<palm_palette_size(); i++) {
		set_pal_entry(i, pal[i].r, pal[i].g, pal[i].b);
	}
	colrange = palm_color_range();

	return 0;
}

void cleanup(void)
{
	free_texture(tex);
	free(fbuf);
	set_video_mode(0x3);
}

void display(void)
{
	static unsigned long msec;

	unsigned int i;
	unsigned char voffs = msec >> 3;
	unsigned char uoffs = msec >> 6;

	unsigned char *fbptr = fbuf;
	unsigned long *uptr = umap;
	unsigned long *vptr = vmap;

	for(i=0; i<64000; i++) {
		unsigned long u = *uptr++;
		unsigned long v = *vptr++;
		unsigned long tx = ((((unsigned long)(u - uoffs) << 3) & 0xff) * tex->width) >> 8;
		unsigned long ty = (((unsigned long)(v + voffs) & 0xff) * tex->height) >> 8;

		unsigned long base = tex->pixels[ty * tex->width + tx];
		long zcue_shift = colrange - (v >> 6);
		if(zcue_shift < 0) {
			zcue_shift = 0;
		}
		*fbptr++ = (unsigned char)(base + zcue_shift);
	}

	copy_frame(fbuf);
	msec += 33;
}

int calc_tunnel_mapping(void)
{
	int i, j;
	unsigned long *uptr, *vptr;
	float diag_dist = sqrt(1.33333 * 1.33333 + 1.0);

	if(!(umap = malloc(WIDTH * HEIGHT * sizeof *umap))) {
		return -1;
	}
	if(!(vmap = malloc(WIDTH * HEIGHT * sizeof *umap))) {
		free(umap);
		return -1;
	}

	uptr = umap;
	vptr = vmap;

	for(i=0; i<HEIGHT; i++) {
		for(j=0; j<WIDTH; j++) {
			float x = ((float)j / WIDTH * 2.0 - 1.0) * 1.33333;
			float y = (float)i / HEIGHT * 2.0 - 1.0;
			float angle, z, dist;

			if(fabs(x) > 0.00001) {
				angle = atan2(y, x) + M_PI;
			} else {
				angle = y < 0.0 ? M_PI / 2.0 : 3.0 * M_PI / 2.0;
			}
			dist = sqrt(x * x + y * y);
			z = 2.0 / dist;

			*uptr++ = (unsigned int)(angle * 0.5 / M_PI * 255.0);
			*vptr++ = (unsigned int)(z * 255.0);
		}
	}
	return 0;
}
